;;;; geo-globe.scm  -*- Scheme -*-
;;;; Kon Lovett, Jul '18
;;;; Kon Lovett, Oct '17

(module geo-globe

(;export
  ;
  make-globe
  globe? check-globe error-globe
  globe-radius-kilometers
  globe-flattening-factor
  ;
  spherical-surface-distance
  great-circle-distance great-circle-distance-radians
  approximate-ellipsoid-distance
  ;
  great-circle-azimuth
  ;
  great-circle-position)

(import scheme
  (chicken base)
  (chicken type)
  (only type-checks-basic define-check+error-type)
  (only (type-checks-numbers scheme) check-real)
  geopoint
  (prefix geo-utils utility-))

(include-relative "geo-utils.types")

(: *make-globe (real real --> globe))
(: globe? (* -> boolean : globe))
(: *globe-radius-kilometers (globe --> real))
(: *globe-flattening-factor (globe --> real))
(: make-globe (real real --> globe))
(: globe-radius-kilometers (globe --> real))
(: globe-flattening-factor (globe --> real))
(: spherical-surface-distance (globe geopoint geopoint --> real))
(: approximate-ellipsoid-distance (globe geopoint geopoint --> real))
(: great-circle-distance (globe geopoint geopoint --> real))
(: great-circle-distance-radians (globe geopoint geopoint --> real))
(: great-circle-azimuth (geopoint geopoint #!rest (list real) --> real))
(: great-circle-position (globe geopoint real real --> geopoint))

;;

(define-record-type globe
  (*make-globe rad flt)
  globe?
  (rad *globe-radius-kilometers)
  (flt *globe-flattening-factor) )

(define-check+error-type globe)

(define (make-globe rad flt)
  (*make-globe
    (check-real 'make-globe rad 'radius)
    (check-real 'make-globe flt 'falttening)) )

(define (globe-radius-kilometers glob)
  (*globe-radius-kilometers (check-globe 'globe-radius-kilometers glob 'globe)))

(define (globe-flattening-factor glob)
  (*globe-flattening-factor (check-globe 'globe-flattening-factor glob 'globe)))

;;

(define (spherical-surface-distance glob gp1 gp2)
  (check-geopoint 'spherical-surface-distance gp1)
  (check-geopoint 'spherical-surface-distance gp2)
  (utility-spherical-surface-distance
    (geopoint-latitude gp1) (geopoint-longitude gp1)
    (geopoint-latitude gp2) (geopoint-longitude gp2)
    (*globe-radius-kilometers (check-globe 'spherical-surface-distance glob 'globe))) )

(define (approximate-ellipsoid-distance glob gp1 gp2)
  (check-globe 'approximate-ellipsoid-distance glob 'globe)
  (check-geopoint 'approximate-ellipsoid-distance gp1)
  (check-geopoint 'approximate-ellipsoid-distance gp2)
  (utility-approximate-ellipsoid-distance
    (geopoint-latitude gp1) (geopoint-longitude gp1)
    (geopoint-latitude gp2) (geopoint-longitude gp2)
    (*globe-radius-kilometers glob)
    (*globe-flattening-factor glob)) )

(define (great-circle-distance glob gp1 gp2)
  (check-geopoint 'great-circle-distance gp1)
  (check-geopoint 'great-circle-distance gp2)
  (utility-great-circle-distance
    (geopoint-latitude gp1) (geopoint-longitude gp1)
    (geopoint-latitude gp2) (geopoint-longitude gp2)
    (*globe-radius-kilometers (check-globe 'great-circle-distance glob 'globe))) )

(define (great-circle-distance-radians glob gp1 gp2)
  (check-geopoint 'great-circle-distance-radians gp1)
  (check-geopoint 'great-circle-distance-radians gp2)
  (utility-great-circle-distance-radians
    (geopoint-latitude gp1) (geopoint-longitude gp1)
    (geopoint-latitude gp2) (geopoint-longitude gp2)
    (*globe-radius-kilometers (check-globe 'great-circle-distance-radians glob 'globe))) )

(define (great-circle-azimuth gp1 gp2 . args)
  (check-geopoint 'great-circle-azimuth gp1)
  (check-geopoint 'great-circle-azimuth gp2)
  (let ((prec (optional args 5)))
    (utility-great-circle-azimuth
      (geopoint-latitude gp1) (geopoint-longitude gp1)
      (geopoint-latitude gp2) (geopoint-longitude gp2)
      (check-real 'great-circle-azimuth prec 'precision)) ) )

(define (great-circle-position glob gp dis azi)
  (check-geopoint 'great-circle-position gp)
  (apply
    make-geopoint
    (receive
      (utility-great-circle-position
        (geopoint-latitude gp) (geopoint-longitude gp)
        (check-real 'great-circle-position dis 'distance)
        (check-real 'great-circle-position azi 'azimuth)
        (*globe-radius-kilometers (check-globe 'great-circle-distance-radians glob 'globe))))) )

) ;module geo-globe
